import glob
import json
import inspect
import os
from os.path import (
    abspath,
    basename,
    dirname,
    exists,
    isdir,
    join,
    normpath,
    relpath,
    sep,
    splitext,
)
import pathlib
import posixpath
import shutil
import sysconfig
import toposort
from typing import Any, Dict, List, Optional, Set, Tuple

from urllib.error import HTTPError
import dataclasses

from setuptools import Extension

from cxxheaderparser.options import ParserOptions
from cxxheaderparser import preprocessor


from .download import download_and_extract_zip
from .config.pyproject_toml import PatchInfo, WrapperConfig, Download

from .autowrap.cxxparser import parse_header
from .autowrap.generator_data import GeneratorData, MissingReporter
from .autowrap.writer import WrapperWriter

from .config.autowrap_yml import AutowrapConfigYaml
from .config.dev_yml import get_dev_config
from .config.pyproject_toml import WrapperConfig, Download

# TODO: eventually provide native preprocessor by default and allow it
#       to be enabled/disabled per-file just in case
if os.getenv("RPYBUILD_PP_GCC") == "1":
    # GCC preprocessor can be 10x faster than pcpp for very complex files
    def make_preprocessor(*args, **kwargs):
        return preprocessor.make_gcc_preprocessor(print_cmd=False, *args, **kwargs)

else:
    make_preprocessor = preprocessor.make_pcpp_preprocessor


class Wrapper:
    """
    Wraps downloading bindings and generating them
    """

    # Used during preprocessing
    # -> should we change this based on what flags the compiler supports?
    _cpp_version = "__cplusplus 201703L"

    def __init__(self, package_name, cfg: WrapperConfig, setup, wwriter: WrapperWriter):
        self.package_name = package_name
        self.cfg = cfg
        self.wwriter = wwriter

        self.setup_root = setup.root
        self.pypi_package = setup.pypi_package
        self.root = join(setup.root, *package_name.split("."))

        # must match PkgCfg.name
        self.name = cfg.name
        self.static_lib = False

        # Compute the extension name, even if we don't create one
        extname = cfg.extension
        if not extname:
            extname = f"_{cfg.name}"

        # must match PkgCfg.libinit_import
        if cfg.libinit:
            libinit_py = cfg.libinit
            if libinit_py == "__init__.py":
                self.libinit_import = package_name
            else:
                pkg = splitext(libinit_py)[0]
                self.libinit_import = f"{package_name}.{pkg}"
        else:
            libinit_py = f"_init{extname}.py"
            self.libinit_import = f"{package_name}._init{extname}"

        self.libinit_import_py = join(self.root, libinit_py)

        self.platform = setup.platform
        self.pkgcfg = setup.pkgcfg

        # Used by pkgcfg
        self.depends = self.cfg.depends

        # Files that are generated AND need to be in the final wheel. Used by build_py
        self.additional_data_files: List[str] = []

        self._all_deps = None

        self._gen_includes = []

        self.extension = None
        if self.cfg.sources or self.cfg.autogen_headers:
            define_macros = [("RPYBUILD_MODULE_NAME", extname)] + [
                tuple(d.split(" ")) for d in self.platform.defines
            ]
            define_macros += [tuple(m.split(" ", 1)) for m in self.cfg.pp_defines]

            # extensions just hold data about what to actually build, we can
            # actually modify extensions all the way up until the build
            # really happens
            extname_full = f"{self.package_name}.{extname}"
            self.extension = Extension(
                extname_full,
                self.cfg.sources,
                define_macros=define_macros,
                language="c++",
            )

            # Add self to extension so that build_ext can query it on OSX
            self.extension.rpybuild_wrapper = self

            # Used if the maven download fails
            self.supported_platforms = setup.project.supported_platforms

        if self.cfg.autogen_headers and not self.cfg.generation_data:
            raise ValueError(
                "generation_data must be specified when autogen_headers/generate is specified"
            )

        # Setup an entry point (written during build_clib)
        entry_point = f"{self.cfg.name} = {self.package_name}.pkgcfg"

        setup_kwargs = setup.setup_kwargs
        ep = setup_kwargs.setdefault("entry_points", {})
        ep.setdefault("robotpybuild", []).append(entry_point)

        self.incdir = join(self.root, "include")
        self.rpy_incdir = join(self.root, "rpy-include")

        self.dev_config = get_dev_config(self.name)

        self._update_addl_data_files()

    def _extract_zip_to(self, dl: Download, dst, cache):
        try:
            download_and_extract_zip(dl.url, dst, cache)
        except HTTPError as e:
            # Check for a 404 error and raise an error if the platform isn't supported.
            if e.code != 404:
                raise e
            else:
                platform_dict = dataclasses.asdict(self.platform)

                os = platform_dict["os"]
                arch = platform_dict["arch"]

                is_os_supported = False
                is_arch_supported = False

                for supp_plat in self.supported_platforms:
                    if supp_plat.os is None or supp_plat.os == os:
                        is_os_supported = True
                        if supp_plat.arch is None or supp_plat.arch == arch:
                            is_arch_supported = True

                if not (is_os_supported and is_arch_supported):
                    if arch == "x86":
                        arch = "32-bit"
                    elif arch == "x86-64":
                        arch = "64-bit"

                    if os == "osx":
                        os = "macOS"

                    if not is_os_supported:
                        arch = ""

                    msg_plat = "{}{}{}".format(arch, " " if arch != "" else "", os)

                    err_msg = "{} is not supported on {}!".format(
                        self.pypi_package, msg_plat
                    )

                    raise OSError(err_msg)
                raise e

    def _add_addl_data_file(self, fullpath):
        if not isdir(fullpath):
            self.additional_data_files.append(relpath(fullpath, self.root))

    # pkgcfg interface
    def get_include_dirs(self) -> List[str]:
        includes = [self.incdir, self.rpy_incdir]
        if self.cfg.download:
            for dl in self.cfg.download:
                if dl.extra_includes:
                    includes += [join(self.incdir, inc) for inc in dl.extra_includes]
        for h in self.cfg.extra_includes:
            includes.append(join(self.setup_root, normpath(h)))
        return includes

    def get_library_dirs(self) -> Optional[List[str]]:
        if self.get_library_full_names():
            return [join(self.root, "lib")]
        return []

    def get_library_dirs_rel(self) -> Optional[List[str]]:
        if self.get_library_full_names():
            return ["lib"]
        return []

    def get_library_names(self) -> Optional[List[str]]:
        libs = []
        if self.cfg.download:
            for dl in self.cfg.download:
                if dl.libs:
                    libs += dl.libs
        return libs

    def get_library_full_names(self) -> Optional[List[str]]:
        if not self.cfg.download:
            return []

        dlopen_libnames = self.get_dlopen_library_names()

        libnames_full = []

        for dl in self.cfg.download:
            libext = dl.libexts.get(self.platform.libext, self.platform.libext)
            if dl.libs:
                for lib in dl.libs:
                    if lib not in dlopen_libnames:
                        libnames_full.append(f"{self.platform.libprefix}{lib}{libext}")
            if dl.dlopenlibs:
                libnames_full += [
                    f"{self.platform.libprefix}{lib}{libext}" for lib in dl.dlopenlibs
                ]

        return libnames_full

    def get_dlopen_library_names(self) -> Optional[List[str]]:
        libs = []
        if self.cfg.download:
            for dl in self.cfg.download:
                if dl.dlopenlibs:
                    libs += dl.dlopenlibs
        return libs

    def get_extra_objects(self) -> Optional[List[str]]:
        pass

    def get_type_casters_cfg(self, casters: Dict[str, Dict[str, Any]]) -> None:
        for ccfg in self.cfg.type_casters:
            cfg = {"hdr": ccfg.header}
            if ccfg.default_arg_cast:
                cfg["darg"] = True

            for typ in ccfg.types:
                casters[typ] = cfg

    def _update_addl_data_files(self) -> List[str]:
        headers = set()
        for ccfg in self.cfg.type_casters:
            headers.add(ccfg.header)

        if headers:
            includes = self.get_include_dirs()
            if includes:
                for hdr in headers:
                    for p in includes:
                        fpath = join(p, hdr)
                        if exists(fpath):
                            self._add_addl_data_file(fpath)

    def all_deps(self):
        if self._all_deps is None:
            self._all_deps = self.pkgcfg.get_all_deps(self.name)
        return self._all_deps

    def _all_includes(self, include_rpyb):
        includes = self.get_include_dirs()
        for dep in self.all_deps():
            dep_inc = dep.get_include_dirs()
            if dep_inc:
                includes.extend(dep_inc)
        if include_rpyb:
            includes.extend(self.pkgcfg.get_pkg("robotpy-build").get_include_dirs())
        return includes

    def _generation_search_path(self):
        return [self.root] + self._all_includes(False)

    def _all_library_dirs(self):
        libs = self.get_library_dirs()
        for dep in self.cfg.depends:
            libdirs = self.pkgcfg.get_pkg(dep).get_library_dirs()
            if libdirs:
                libs.extend(libdirs)
        return libs

    def _all_library_names(self):
        libs = list(
            set(self.get_library_names()) | set(self.get_dlopen_library_names())
        )
        for dep in self.cfg.depends:
            pkg = self.pkgcfg.get_pkg(dep)
            libnames = pkg.get_library_names()
            if libnames:
                libs.extend(libnames)
        return list(reversed(libs))

    def _all_extra_objects(self):
        libs = []
        for dep in self.cfg.depends:
            pkg = self.pkgcfg.get_pkg(dep)
            libnames = pkg.get_extra_objects()
            if libnames:
                libs.extend(libnames)
        return list(reversed(libs))

    def _all_casters(self):
        casters = {}
        for dep in self.all_deps():
            dep.get_type_casters_cfg(casters)
        self.pkgcfg.get_pkg("robotpy-build").get_type_casters_cfg(casters)
        self.get_type_casters_cfg(casters)

        # make each configuration unique
        for k, v in list(casters.items()):
            v = v.copy()
            v["typename"] = k
            casters[k] = v

        # add non-namespaced versions of all casters
        # -> in theory this could lead to a conflict, but
        #    let's see how it works in practice?
        for k, v in list(casters.items()):
            k = k.split("::")[-1]
            casters[k] = v
        return casters

    def on_build_dl(self, cache: str, srcdir: str):
        pkgcfgpy = join(self.root, "pkgcfg.py")
        srcdir = join(srcdir, self.name)

        try:
            os.unlink(self.libinit_import_py)
        except OSError:
            pass

        try:
            os.unlink(pkgcfgpy)
        except OSError:
            pass

        libnames_full = []
        all_libs = []
        downloads = self.cfg.download
        if downloads:
            libnames_full, all_libs = self._clean_and_download(downloads, cache, srcdir)

        self._write_libinit_py(libnames_full)
        self._write_pkgcfg_py(pkgcfgpy, libnames_full)

        return all_libs

    def _apply_patches(self, patches: List[PatchInfo], root: str):
        import patch

        for p in patches:
            patch_path = join(self.setup_root, normpath(p.patch))
            ps = patch.PatchSet()
            with open(patch_path, "rb") as fp:
                if not ps.parse(fp):
                    raise ValueError(f"Error parsing patch '{patch_path}'")

            if not ps.apply(strip=p.strip, root=root):
                raise ValueError(f"Error applying patch '{patch_path}' to '{root}'")

    def _clean_and_download(
        self, downloads: List[Download], cache: str, srcdir: str
    ) -> Tuple[List[str], List[str]]:
        libdir = join(self.root, "lib")
        incdir = join(self.root, "include")

        add_libdir = False
        add_incdir = False

        # Remove downloaded/generated artifacts first
        shutil.rmtree(libdir, ignore_errors=True)
        shutil.rmtree(incdir, ignore_errors=True)
        shutil.rmtree(srcdir, ignore_errors=True)

        dlopen_libnames = self.get_dlopen_library_names()
        libnames_full = []
        all_libs = []

        for dl in downloads:
            # extract the whole thing into a directory when using for sources
            if dl.sources is not None:
                download_and_extract_zip(dl.url, srcdir, cache)
                sources = [join(srcdir, normpath(s)) for s in dl.sources]
                self.extension.sources.extend(sources)
                if dl.patches:
                    self._apply_patches(dl.patches, srcdir)
            elif dl.sources is not None:
                raise ValueError("sources must be None if use_sources is False!")
            elif dl.patches is not None:
                raise ValueError("patches must be None if use_sources is False!")

            if dl.libs or dl.dlopenlibs:
                add_libdir = True
                extract_names = []
                os.makedirs(libdir)

                libext = dl.libexts.get(self.platform.libext, self.platform.libext)
                linkext = dl.linkexts.get(self.platform.linkext, self.platform.linkext)
                if dl.libs:
                    for lib in dl.libs:
                        if lib not in dlopen_libnames:
                            name = f"{self.platform.libprefix}{lib}{libext}"
                            libnames_full.append(name)
                            extract_names.append(name)
                            if libext != linkext:
                                extract_names.append(
                                    f"{self.platform.libprefix}{lib}{linkext}"
                                )

                if dl.dlopenlibs:
                    dlopen_libnames_full = [
                        f"{self.platform.libprefix}{lib}{libext}"
                        for lib in dl.dlopenlibs
                    ]
                    libnames_full += dlopen_libnames_full
                    extract_names += dlopen_libnames_full

                to = {
                    posixpath.join(dl.libdir, libname): join(libdir, libname)
                    for libname in extract_names
                }
                all_libs.extend(to.values())
            else:
                to = {}

            if dl.incdir is not None:
                to[dl.incdir] = self.incdir
                add_incdir = True

            download_and_extract_zip(dl.url, to, cache)

            if dl.header_patches:
                self._apply_patches(dl.header_patches, incdir)

        if add_incdir:
            for f in glob.glob(join(glob.escape(incdir), "**"), recursive=True):
                self._add_addl_data_file(f)

        if add_libdir:
            for f in glob.glob(join(glob.escape(libdir), "**"), recursive=True):
                self._add_addl_data_file(f)

        return libnames_full, all_libs

    def _write_libinit_py(self, libnames):
        # This file exists to ensure that any shared library dependencies
        # are loaded for the compiled extension

        init = inspect.cleandoc(
            """
        
        # This file is automatically generated, DO NOT EDIT
        # fmt: off

        from os.path import abspath, join, dirname, exists
        _root = abspath(dirname(__file__))

        ##IMPORTS##

        """
        )

        init += "\n"

        if libnames:
            if self.platform.os == "osx":
                init += "from ctypes import CDLL, RTLD_GLOBAL\n\n"
            else:
                init += "from ctypes import cdll\n\n"

            for libname in libnames:
                init += "try:\n"
                if self.platform.os == "osx":
                    init += f'    _lib = CDLL(join(_root, "lib", "{libname}"), mode=RTLD_GLOBAL)\n'
                else:
                    init += f'    _lib = cdll.LoadLibrary(join(_root, "lib", "{libname}"))\n'

                init += "except FileNotFoundError:\n"
                init += f'    if not exists(join(_root, "lib", "{libname}")):\n'
                init += f'        raise FileNotFoundError("{libname} was not found on your system. Is this package correctly installed?")\n'
                if self.platform.os == "windows":
                    init += f'    raise Exception("{libname} could not be loaded. Do you have Visual Studio C++ Redistributible 2019 installed?")\n\n'
                else:
                    init += f'    raise FileNotFoundError("{libname} could not be loaded. There is a missing dependency.")\n\n'
        imports = []
        for dep in self.cfg.depends:
            pkg = self.pkgcfg.get_pkg(dep)
            if pkg.libinit_import:
                imports.append(pkg.libinit_import)

        if imports:
            imports = "# runtime dependencies\nimport " + "\nimport ".join(imports)
        else:
            imports = ""

        init = init.replace("##IMPORTS##", imports)

        with open(self.libinit_import_py, "w") as fp:
            fp.write(init)

        self._add_addl_data_file(self.libinit_import_py)

    def _write_pkgcfg_py(self, fname, libnames_full):
        library_dirs = "[]"
        library_dirs_rel = []
        library_names = self.get_library_names()
        if library_names:
            library_dirs = '[join(_root, "lib")]'
            library_dirs_rel = ["lib"]

        deps = []
        for dep in self.cfg.depends:
            pkg = self.pkgcfg.get_pkg(dep)
            if not pkg.static_lib:
                deps.append(dep)

        # write pkgcfg.py
        pkgcfg = inspect.cleandoc(
            f"""
        # fmt: off
        # This file is automatically generated, DO NOT EDIT

        from os.path import abspath, join, dirname
        _root = abspath(dirname(__file__))

        libinit_import = "{self.libinit_import}"
        depends = {repr(deps)}
        pypi_package = {repr(self.pypi_package)}

        def get_include_dirs():
            return [join(_root, "include"), join(_root, "rpy-include")##EXTRAINCLUDES##]

        def get_library_dirs():
            return {library_dirs}

        def get_library_dirs_rel():
            return {repr(library_dirs_rel)}
        
        def get_library_names():
            return {repr(library_names)}

        def get_library_full_names():
            return {repr(libnames_full)}
        """
        )

        extraincludes = ""
        if self.cfg.extra_includes:
            # these are relative to the root of the project, need
            # to resolve the path relative to the package
            pth = join(*self.package_name.split("."))

            for h in self.cfg.extra_includes:
                h = '", "'.join(relpath(normpath(h), pth).split(sep))
                extraincludes += f', join(_root, "{h}")'

        pkgcfg = pkgcfg.replace("##EXTRAINCLUDES##", extraincludes)

        type_casters = {}
        self.get_type_casters_cfg(type_casters)
        if type_casters:
            pkgcfg += "\n\n"
            pkgcfg += inspect.cleandoc(
                f"""

            def get_type_casters_cfg(casters):
                casters.update({repr(type_casters)})

            def get_type_casters(casters):
                t = {{}}
                get_type_casters_cfg(t)
                for k, v in t.items():
                    if "hdr" in v:
                        casters[k] = v["hdr"]
            """
            )

        with open(fname, "w") as fp:
            fp.write(pkgcfg)

        self._add_addl_data_file(fname)

    def on_build_gen(
        self, cxx_gen_dir, missing_reporter: Optional[MissingReporter] = None
    ):
        if not self.cfg.autogen_headers:
            return

        cxx_gen_dir = join(cxx_gen_dir, self.name)

        if missing_reporter:
            report_only = True
        else:
            report_only = False
            missing_reporter = MissingReporter()

        hppoutdir = join(self.rpy_incdir, "rpygen")

        pp_includes = self._all_includes(True) + [sysconfig.get_path("include")]

        # TODO: only regenerate files if the generated files
        #       have changed
        if not report_only:
            if self.dev_config.only_generate is None:
                shutil.rmtree(cxx_gen_dir, ignore_errors=True)
                shutil.rmtree(hppoutdir, ignore_errors=True)

            os.makedirs(cxx_gen_dir, exist_ok=True)
            os.makedirs(hppoutdir, exist_ok=True)

        per_header = False
        data_fname = self.cfg.generation_data
        if self.cfg.generation_data:
            datapath = join(self.setup_root, normpath(self.cfg.generation_data))
            per_header = isdir(datapath)
            if not per_header:
                data = AutowrapConfigYaml.from_file(datapath)
        else:
            data = AutowrapConfigYaml()

        pp_defines = [self._cpp_version] + self.platform.defines + self.cfg.pp_defines
        casters = self._all_casters()

        # These are written to file to make it easier for dev mode to work
        classdeps = {}

        if self.dev_config.only_generate is not None:
            only_generate = {n: True for n in self.dev_config.only_generate}
        else:
            only_generate = None

        generation_search_path = self._generation_search_path()

        for name, header in self.cfg.autogen_headers.items():
            header = normpath(header)
            for path in generation_search_path:
                header_path = join(path, header)
                if exists(header_path):
                    header_root = pathlib.Path(path)
                    break
            else:
                import pprint

                pprint.pprint(generation_search_path)
                raise ValueError("could not find " + header)

            if not report_only:
                classdeps_dst = join(cxx_gen_dir, f"{name}.json")
                classdeps[name] = classdeps_dst

            if per_header:
                data_fname = join(datapath, name + ".yml")
                if not exists(data_fname):
                    print("WARNING: could not find", data_fname)
                    data = AutowrapConfigYaml()
                else:
                    data = AutowrapConfigYaml.from_file(data_fname)

            if only_generate is not None and not only_generate.pop(name, False):
                continue

            popts = ParserOptions(
                preprocessor=make_preprocessor(
                    defines=pp_defines,
                    include_paths=pp_includes,
                    encoding=data.encoding,
                )
            )

            gendata = GeneratorData(data)

            try:
                hctx = parse_header(
                    name,
                    pathlib.Path(header_path),
                    header_root,
                    gendata,
                    popts,
                    casters,
                    report_only,
                )

                if not report_only:
                    generated_sources = self.wwriter.write_files(
                        hctx, name, cxx_gen_dir, hppoutdir, classdeps_dst
                    )
                    self.extension.sources.extend(
                        [relpath(src, self.setup_root) for src in generated_sources]
                    )
            except Exception as e:
                raise ValueError(f"processing {header}") from e

            gendata.report_missing(data_fname, missing_reporter)

        if only_generate:
            unused = ", ".join(sorted(only_generate))
            # raise ValueError(f"only_generate specified unused headers! {unused}")
            # TODO: make this a warning

        if not report_only:
            for name, contents in missing_reporter.as_yaml():
                print("WARNING: some items not in generation yaml for", basename(name))
                print(contents)

        # generate an inline file that can be included + called
        if not report_only:
            self._write_wrapper_hpp(cxx_gen_dir, classdeps)
            gen_includes = [cxx_gen_dir]
        else:
            gen_includes = []

        self._gen_includes = gen_includes

        for f in glob.glob(join(glob.escape(hppoutdir), "*.hpp")):
            self._add_addl_data_file(f)

    def finalize_extension(self):
        if self.extension is None:
            return

        # Add the root to the includes (but only privately)
        root_includes = [self.root]

        # update the build extension so that build_ext works
        # use normpath to get rid of .. otherwise gcc is weird
        self.extension.include_dirs = [
            normpath(p)
            for p in (self._all_includes(True) + self._gen_includes + root_includes)
        ]
        self.extension.library_dirs = self._all_library_dirs()
        self.extension.libraries = self._all_library_names()
        self.extension.extra_objects = self._all_extra_objects()

    def _write_wrapper_hpp(self, outdir, classdeps):
        decls = []
        begin_calls = []
        finish_calls = []

        # Need to ensure that wrapper initialization is called in base order
        # so we have to toposort it here. The data is written at gen time
        # to JSON files
        types2name = {}
        types2deps = {}
        ordering = []

        for name, jsonfile in classdeps.items():
            with open(jsonfile) as fp:
                dep = json.load(fp)

            # make sure objects without classes are also included!
            if not dep:
                ordering.append(name)

            for clsname, bases in dep.items():
                if clsname in types2name:
                    raise ValueError(f"{name} ({jsonfile}): duplicate class {clsname}")
                types2name[clsname] = name
                types2deps[clsname] = bases[:]

        to_sort: Dict[str, Set[str]] = {}
        for clsname, bases in types2deps.items():
            clsname = types2name[clsname]
            deps = to_sort.setdefault(clsname, set())
            for base in bases:
                base = types2name.get(base)
                if base and base != clsname:
                    deps.add(base)

        ordering.extend(toposort.toposort_flatten(to_sort, sort=True))

        for name in ordering:
            decls.append(f"void begin_init_{name}(py::module &m);")
            decls.append(f"void finish_init_{name}();")
            begin_calls.append(f"    begin_init_{name}(m);")
            finish_calls.append(f"    finish_init_{name}();")

        content = (
            inspect.cleandoc(
                """

        // This file is autogenerated, DO NOT EDIT
        #pragma once
        #include <robotpy_build.h>

        // forward declarations
        ##DECLS##

        static void initWrapper(py::module &m) {
        ##BEGIN_CALLS##

        ##FINISH_CALLS##
        }
        
        """
            )
            .replace("##DECLS##", "\n".join(decls))
            .replace("##BEGIN_CALLS##", "\n".join(begin_calls))
            .replace("##FINISH_CALLS##", "\n".join(finish_calls))
        )

        with open(join(outdir, "rpygen_wrapper.hpp"), "w") as fp:
            fp.write(content)



import importlib.util
import os
from os.path import abspath, exists, dirname, join

from setuptools import Command

from .util import get_install_root


class _BuiltEnv(Command):

    user_options = [("build-lib=", "d", 'directory to "build" (copy) to')]

    def initialize_options(self):
        self.build_lib = None

    def finalize_options(self):
        self.set_undefined_options("build", ("build_lib", "build_lib"))

    def setup_built_env(self):

        # Gather information for n
        data = {"mapping": {}}

        # OSX-specific: need to set DYLD_LIBRARY_PATH otherwise modules don't
        # work. Luckily, that information was computed when building the
        # extensions...
        env = os.environ.copy()
        dyld_path = set()

        # Requires information from build_ext to work
        build_ext = self.get_finalized_command("build_ext")
        if build_ext.inplace:
            data["out"] = get_install_root(self)
        else:
            data["out"] = self.build_lib

        # Ensure that the associated packages can always be found locally
        for wrapper in build_ext.wrappers:
            pkgdir = wrapper.package_name.split(".")
            init_py = abspath(join(self.build_lib, *pkgdir, "__init__.py"))
            if exists(init_py):
                data["mapping"][wrapper.package_name] = init_py

        # Ensure that the built extension can always be found
        build_ext.resolve_libs()
        for ext in build_ext.extensions:
            fname = build_ext.get_ext_filename(ext.name)
            data["mapping"][ext.name] = abspath(join(self.build_lib, fname))

            rpybuild_libs = getattr(ext, "rpybuild_libs", None)
            if rpybuild_libs:
                for pth, _ in rpybuild_libs.values():
                    dyld_path.add(dirname(pth))

        # OSX-specific
        if dyld_path:
            dyld_path = ":".join(dyld_path)
            if "DYLD_LIBRARY_PATH" in env:
                dyld_path += ":" + env["DYLD_LIBRARY_PATH"]
            env["DYLD_LIBRARY_PATH"] = dyld_path

        return data, env


class _PackageFinder:
    """
    Custom loader to allow loading built modules from their location
    in the build directory (as opposed to their install location)
    """

    # Set this to mapping returned from _BuiltEnv.setup_built_env
    mapping = {}

    @classmethod
    def find_spec(cls, fullname, path, target=None):
        m = cls.mapping.get(fullname)
        if m:
            return importlib.util.spec_from_file_location(fullname, m)



import json
import os
import subprocess
import sys
import typing

try:
    from setuptools.errors import BaseError
except ImportError:
    from distutils.errors import DistutilsError as BaseError

from ._built_env import _BuiltEnv, _PackageFinder


class UpdateInitError(BaseError):
    pass


class UpdateInit(_BuiltEnv):
    update_list: typing.List[str]

    command_name = "update_init"
    description = (
        "Updates __init__.py files using settings from tool.robotpy-build.update_init"
    )

    def run(self):
        # cannot use when cross-compiling
        if (
            "_PYTHON_HOST_PLATFORM" in os.environ
            or "PYTHON_CROSSENV" in os.environ
            or not self.update_list
        ):
            return

        data, env = self.setup_built_env()
        data["update_list"] = self.update_list

        data_json = json.dumps(data)

        # Execute in a subprocess in case it crashes
        args = [sys.executable, "-m", __name__]
        try:
            subprocess.run(args, input=data_json.encode("utf-8"), env=env, check=True)
        except subprocess.CalledProcessError:
            raise UpdateInitError(
                "Failed to generate .pyi file (see above, or set RPYBUILD_SKIP_PYI=1 to ignore) via %s"
                % (args,)
            ) from None


def main():
    cfg = json.load(sys.stdin)

    # Configure custom loader
    _PackageFinder.mapping = cfg["mapping"]
    sys.meta_path.insert(0, _PackageFinder)

    from .. import tool

    # Update init

    for to_update in cfg["update_list"]:

        sys.argv = ["<dummy>", "create-imports", "-w"] + to_update.split(" ", 1)

        retval = tool.main()
        if retval != 0:
            break


if __name__ == "__main__":
    main()



from rpytest import ft


def test_another_diamond_b():
    o = ft.MVB()
    assert o.b == 1

    assert o.get_b_b() == 1


def test_another_diamond_c():
    o = ft.MVC()
    assert o.b == 1
    assert o.c == 2

    assert o.get_b_b() == 1
    assert o.get_c_b() == 1

    assert o.get_c_c() == 2


def test_another_diamond_d0():
    o = ft.MVD0()
    assert o.b == 1
    assert o.c == 2
    assert o.d0 == 3

    assert o.get_b_b() == 1
    assert o.get_c_b() == 1
    assert o.get_d0_b() == 1

    assert o.get_c_c() == 2
    assert o.get_d0_c() == 2

    assert o.get_d0_d0() == 3


def test_another_diamond_d1():
    o = ft.MVD1()
    assert o.b == 1
    assert o.c == 2
    assert o.d1 == 4

    assert o.get_b_b() == 1
    assert o.get_c_b() == 1
    assert o.get_d1_b() == 1

    assert o.get_c_c() == 2
    assert o.get_d1_c() == 2

    assert o.get_d1_d1() == 4


def test_another_diamond_e():
    o = ft.MVE()
    assert o.b == 1
    assert o.c == 2
    assert o.d0 == 3
    assert o.d1 == 4
    assert o.e == 5

    assert o.get_b_b() == 1
    assert o.get_c_b() == 1
    assert o.get_d0_b() == 1
    assert o.get_d1_b() == 1
    assert o.get_e_b() == 1

    assert o.get_c_c() == 2
    assert o.get_d0_c() == 2
    assert o.get_d1_c() == 2
    assert o.get_e_c() == 2

    assert o.get_d0_d0() == 3
    assert o.get_e_d0() == 3

    assert o.get_d1_d1() == 4
    assert o.get_e_d1() == 4

    assert o.get_e_e() == 5


def test_another_diamond_f():
    o = ft.MVF()
    assert o.b == 1
    assert o.c == 2
    assert o.d0 == 3
    assert o.d1 == 4
    assert o.e == 5
    assert o.f == 6

    assert o.get_b_b() == 1
    assert o.get_c_b() == 1
    assert o.get_d0_b() == 1
    assert o.get_d1_b() == 1
    assert o.get_e_b() == 1
    assert o.get_f_b() == 1

    assert o.get_c_c() == 2
    assert o.get_d0_c() == 2
    assert o.get_d1_c() == 2
    assert o.get_e_c() == 2
    assert o.get_f_c() == 2

    assert o.get_d0_d0() == 3
    assert o.get_e_d0() == 3
    assert o.get_f_d0() == 3

    assert o.get_d1_d1() == 4
    assert o.get_e_d1() == 4
    assert o.get_f_d1() == 4

    assert o.get_e_e() == 5
    assert o.get_f_e() == 5

    assert o.get_f_f() == 6



from rpytest import ft
import pytest


class PyGChild(ft.PChild):
    pass


def test_ft_protection():
    # can create base
    b = ft.PBase()

    assert b.getChannel() == 9

    # protected things are available with leading underscore
    b._setChannel(7)
    assert b.getChannel() == 7

    # child with protected constructor is usable
    c = ft.PChild(1)
    assert c.getChannel() == 1

    # Inheritance of child works
    pyc = PyGChild(2)
    assert pyc.getChannel() == 2
    pyc._setChannel(5)
    assert pyc.getChannel() == 5

    # can create grandchild
    gc = ft.PGChild(3)
    assert gc.getChannel() == 3

    # isinstance works too
    assert isinstance(c, ft.PBase)
    assert isinstance(pyc, ft.PBase)
    assert isinstance(gc, ft.PBase)

    # private final should not fail to compile, and children
    # should still work
    # .. oddly enough the method doesn't get hidden, must be
    #    because of pybind11's inheritance. Probably should fix?

    assert b.privateFinalTestC() == 1
    assert b.privateFinalTestGC() == 10
    assert b.privateOverrideTestC() == 100

    assert c.privateFinalTestC() == 2
    assert c.privateFinalTestGC() == 20
    assert c.privateOverrideTestC() == 200

    # Well, this is an unexpected result, but it works too
    assert gc.privateFinalTestC() == 2
    assert gc.privateFinalTestGC() == 30
    assert gc.privateOverrideTestC() == 200



# just ensures the module built correctly, nothing fancy
import rpytest.dl


def test_downloaded_fn():
    assert rpytest.dl.downloaded_fn(3) == 0x42 + 3


def test_extra_content():
    assert rpytest.dl.extra_content() == True



from rpytest import ft
import pytest


def test_array_out():
    assert ft.fnParamArrayOut() == (4, [5, 6, 7])


def test_array_out_w_default():
    assert ft.fnParamArrayOutWithDefault() == (4, [0, 6, 2])


def test_fund_ptr():
    assert ft.fnParamFundPtr(4) == (3, 5)


def test_fund_ref():
    assert ft.fnParamFundRef(6) == (7, 5)


def test_fund_const_ref():
    assert ft.fnParamFundConstRef(1, 2) == 3


def test_fn_disable_none():
    with pytest.raises(TypeError):
        ft.fnParamDisableNone(None)

    assert ft.fnParamDisableNone(ft.Param())


def test_fn_disable_all_none():
    with pytest.raises(TypeError):
        ft.fnParamDisableAllNone(None, None)

    with pytest.raises(TypeError):
        ft.fnParamDisableAllNone(None, ft.Param())

    with pytest.raises(TypeError):
        ft.fnParamDisableAllNone(ft.Param(), None)

    assert ft.fnParamDisableAllNone(ft.Param(), ft.Param())


def _callable():
    pass


def test_fn_auto_disable_none():
    with pytest.raises(TypeError):
        ft.fnParamAutoDisableNone(None)

    assert ft.fnParamAutoDisableNone(_callable)


def test_fn_allow_none():
    assert ft.fnParamAllowNone(_callable)
    assert ft.fnParamAllowNone(None) == False


def test_fn_disable_default():
    assert ft.fnParamDisableDefault(1) == 2
    with pytest.raises(TypeError):
        ft.fnParamDisableDefault()



#
# Defines data that is consumed by the header2whatever hooks/templates
# to modify the generated files
#

import enum
from typing import Dict, List, Tuple, Optional

from pydantic import validator, Field
from .util import Model, _generating_documentation
import yaml


class ParamData(Model):
    """Various ways to modify parameters"""

    #: Set parameter name to this
    name: Optional[str] = None

    #: Change C++ type emitted
    x_type: Optional[str] = None

    #: Default value for parameter
    default: Optional[str] = None

    #: Disable default value
    no_default: Optional[bool] = False

    #: Disallow implicit conversions from None. This defaults to True for built
    #: in types and types that are obviously std::function (does not handle all
    #: cases, in which case this should be explicitly specified)
    disable_none: Optional[bool] = None

    #: Disables a default cast caused by ``default_arg_cast``
    disable_type_caster_default_cast: bool = False

    #: Force this to be an 'out' parameter
    #:
    #: .. seealso:: :ref:`autowrap_out_params`
    #:
    force_out: bool = False

    #: Force an array size
    array_size: Optional[int] = None

    #: Ignore this parameter
    ignore: bool = False


class BufferType(str, enum.Enum):
    #: The buffer must indicate that it is readable (such as bytes, or bytearray)
    IN = "in"

    #: The buffer must indicate that it is writeable (such as a bytearray)
    OUT = "out"

    #: The buffer must indicate that it readable or writeable (such as a bytearray)
    INOUT = "inout"


class BufferData(Model):
    #: Indicates what type of python buffer is required
    type: BufferType

    #: Name of C++ parameter that the buffer will use
    src: str

    #: Name of the C++ length parameter. An out-only parameter, it will be set
    #: to the size of the python buffer, and will be returned so the caller can
    #: determine how many bytes were written
    len: str

    #: If specified, the minimum size of the python buffer
    minsz: Optional[int] = None


class ReturnValuePolicy(enum.Enum):
    """
    See `pybind11 documentation <https://pybind11.readthedocs.io/en/stable/advanced/functions.html#return-value-policies>`_
    for what each of these values mean.
    """

    TAKE_OWNERSHIP = "take_ownership"
    COPY = "copy"
    MOVE = "move"
    REFERENCE = "reference"
    REFERENCE_INTERNAL = "reference_internal"
    AUTOMATIC = "automatic"
    AUTOMATIC_REFERENCE = "automatic_reference"


class FunctionData(Model):
    """
    Customize the way the autogenerator binds a function.

    .. code-block:: yaml

       functions:
         # for non-overloaded functions, just specify the name + customizations
         name_of_non_overloaded_fn:
           # add customizations for function here

         # For overloaded functions, specify the name, but each overload
         # separately
         my_overloaded_fn:
           overloads:
             int, int:
               # customizations for `my_overloaded_fn(int, int)`
             int, int, int:
               # customizations for `my_overloaded_fn(int, int, int)`
    """

    #: If True, don't wrap this
    ignore: bool = False

    #: If True, don't wrap this, but provide a pure virtual implementation
    ignore_pure: bool = False

    #: Most of the time, you will want to specify ``ignore`` instead.
    #:
    #: If True, don't expose this function to python. If a trampoline is supposed
    #: to be generated, it will still be generated. You will likely want to use
    #: ``trampoline_cpp_code`` if you specify this.
    ignore_py: bool = False

    #: Generate this in an `#ifdef`
    ifdef: Optional[str] = None
    #: Generate this in an `#ifndef`
    ifndef: Optional[str] = None

    #: Use this code instead of the generated code
    cpp_code: Optional[str] = None

    #: Docstring for the function, will attempt to convert Doxygen docs if omitted
    doc: Optional[str] = None

    #: Text to append to the (autoconverted) docstring for the function
    doc_append: Optional[str] = None

    #: Disallow implicit conversions from None for all parameters. See also
    #: ``disable_none`` in ParamData.
    disable_none: Optional[bool] = None

    #: If True, prepends an underscore to the python name
    internal: bool = False

    #: Use this to set the name of the function as exposed to python
    #:
    #: When applied to a constructor, creates a static method for the overload instead.
    rename: Optional[str] = None

    #: Mechanism to override individual parameters
    param_override: Dict[str, ParamData] = {}

    #: If specified, put the function in a sub.pack.age
    subpackage: Optional[str] = None

    #: By default, robotpy-build will release the GIL whenever a wrapped
    #: function is called.
    no_release_gil: Optional[bool] = None

    buffers: List[BufferData] = []

    overloads: Dict[str, "FunctionData"] = {}

    #: Adds py::keep_alive<x,y> to the function. Overrides automatic
    #: keepalive support, which retains references passed to constructors.
    #: https://pybind11.readthedocs.io/en/stable/advanced/functions.html#keep-alive
    keepalive: Optional[List[Tuple[int, int]]] = None

    #: https://pybind11.readthedocs.io/en/stable/advanced/functions.html#return-value-policies
    return_value_policy: ReturnValuePolicy = ReturnValuePolicy.AUTOMATIC

    #: If this is a function template, this is a list of instantiations
    #: that you wish to provide. This is a list of lists, where the inner
    #: list is the template parameters for that function
    template_impls: Optional[List[List[str]]] = None

    #: Specify custom C++ code for the virtual function trampoline
    trampoline_cpp_code: Optional[str] = None

    #: Specify a transformation lambda to be used when this virtual function
    #: is called from C++. This inline code should be a lambda that has the same
    #: arguments as the original C++ virtual function, except the first argument
    #: will be a py::function with the python overload
    #:
    #: cpp_code should also be specified for this to be useful
    #:
    #: For example, to transform a function that takes an iostream into a function
    #: that returns a string:
    #:
    #: .. code-block:: yaml
    #:
    #:    cpp_code: |
    #:      [](MyClass* self) {
    #:        return "string";
    #:      }
    #:    virtual_xform: |
    #:      [](py::function fn, MyClass* self, std::iostream &is) {
    #:         std::string d = py::cast(fn());
    #:         is << d;
    #:      }
    #:
    virtual_xform: Optional[str] = None

    @validator("overloads", pre=True)
    def validate_overloads(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = FunctionData()
        return value

    @validator("virtual_xform")
    def validate_virtual_xform(cls, v, values):
        if v and values.get("trampoline_cpp_code"):
            raise ValueError(
                "cannot specify trampoline_cpp_code and virtual_xform for the same method"
            )
        return v


if not _generating_documentation:
    FunctionData.update_forward_refs()


class PropAccess(enum.Enum):
    """Whether a property is writable."""

    #: Determine read/read-write automatically:
    #:
    #: * If a struct/union, default to readwrite
    #: * If a class, default to readwrite if a basic type that isn't a
    #:   reference, otherwise default to readonly
    AUTOMATIC = "auto"

    #: Allow python users access to the value, but ensure it can't
    #: change. This is useful for properties that are defined directly
    #: in the class
    READONLY = "readonly"

    #: Allows python users to read/write the value
    READWRITE = "readwrite"


class PropData(Model):
    #: If set to True, this property is not made available to python
    ignore: bool = False

    #: Set the python name of this property to the specified string
    rename: Optional[str] = None

    #: Python code access to this property
    access: PropAccess = PropAccess.AUTOMATIC

    #: Docstring for the property (only available on class properties)
    doc: Optional[str] = None

    #: Text to append to the (autoconverted) docstring
    doc_append: Optional[str] = None


class EnumValue(Model):
    #: If set to True, this property is not made available to python
    ignore: bool = False

    #: Set the python name of this enum value to the specified string
    rename: Optional[str] = None

    #: Docstring for the enum value
    doc: Optional[str] = None

    #: Text to append to the (autoconverted) docstring
    doc_append: Optional[str] = None


class EnumData(Model):
    #: Set your own docstring for the enum
    doc: Optional[str] = None

    #: Text to append to the (autoconverted) docstring
    doc_append: Optional[str] = None

    #: If set to True, this property is not made available to python
    ignore: bool = False

    #: Set the python name of this enum to the specified string
    rename: Optional[str] = None

    #: Remove this prefix from autogenerated enum value name
    value_prefix: Optional[str] = None

    #: If specified, put the enum in a sub.pack.age (ignored for
    #: enums that are part of classes)
    subpackage: Optional[str] = None

    values: Dict[str, EnumValue] = {}

    #: This will insert code right before the semicolon ending the enum py
    #: definition. You can use this to easily insert additional custom values
    #: without using the global inline_code mechanism.
    inline_code: Optional[str] = None

    #: Tell pybind11 to create an enumeration that also supports rudimentary
    #: arithmetic and bit-level operations like comparisons, and, or, xor,
    #: negation, etc.
    arithmetic: bool = False


class ClassData(Model):
    #: Docstring for the class
    doc: Optional[str] = None

    #: Text to append to the (autoconverted) docstring
    doc_append: Optional[str] = None

    ignore: bool = False

    #: List of bases to ignore. Name must include any template specializations.
    ignored_bases: List[str] = []

    #: Specify fully qualified names for the bases. If the base has a template
    #: parameter, you must include it. Only needed if it can't be automatically
    #: detected directly from the text.
    base_qualnames: Dict[str, str] = {}

    attributes: Dict[str, PropData] = {}
    enums: Dict[str, EnumData] = {}
    methods: Dict[str, FunctionData] = {}

    is_polymorphic: Optional[bool] = None
    force_no_trampoline: bool = False
    force_no_default_constructor: bool = False

    #: pybind11 will detect multiple inheritance automatically if a
    #: class directly derives from multiple classes. However,
    #: If the class derives from classes that participate in multiple
    #: inheritance, pybind11 won't detect it automatically, so this
    #: flag is needed.
    force_multiple_inheritance: bool = False

    #: If there are circular dependencies, this will help you resolve them
    #: manually. TODO: make it so we don't need this
    force_depends: List[str] = []

    #: Use this to bring in type casters for a particular type that may have
    #: been hidden (for example, with a typedef or definition in another file),
    #: instead of explicitly including the header. This should be the full
    #: namespace of the type.
    force_type_casters: List[str] = []

    #: If the object shouldn't be deleted by pybind11, use this. Disables
    #: implicit constructors.
    nodelete: bool = False

    #: Set the python name of the class to this
    rename: Optional[str] = None

    #: This is deprecated and has no effect
    shared_ptr: bool = True

    #: If specified, put the class in a sub.pack.age. Ignored
    #: for functions attached to a class. When template parameters
    #: are used, must define subpackage on template instances
    #: instead
    subpackage: Optional[str] = None

    #: Extra 'using' directives to insert into the trampoline and the
    #: wrapping scope
    typealias: List[str] = []

    #: Fully-qualified pre-existing constant that will be inserted into the
    #: trampoline and wrapping scopes as a constexpr
    constants: List[str] = []

    #: If this is a template class, a list of the parameters if it can't
    #: be autodetected (currently can't autodetect). If there is no space
    #: in the parameter, then it is assumed to be a 'typename', otherwise
    #: the parameter is split by space and the first item is the type and
    #: the second parameter is the name (useful for integral templates)
    template_params: Optional[List[str]] = None

    #: If this is a template class, the specified C++ code is inserted
    #: into the template definition
    template_inline_code: str = ""

    #: If this class has an associated trampoline, add this code inline at
    #: the bottom of the trampoline class. This is rarely useful.
    trampoline_inline_code: Optional[str] = None

    #: This will insert code right before the semicolon ending the class py
    #: definition. You can use this to easily insert additional custom functions
    #: without using the global inline_code mechanism.
    #:
    #: .. code-block: C++
    #:
    #:   cls_MyClass
    #:      .def("something", &MyClass::Something)  // autogenerated
    #:      // inline code is inserted here
    #:      ;
    #:
    inline_code: Optional[str] = None

    @validator("attributes", pre=True)
    def validate_attributes(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = PropData()
        return value

    @validator("enums", pre=True)
    def validate_enums(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = EnumData()
        return value

    @validator("methods", pre=True)
    def validate_methods(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = FunctionData()
        return value


class TemplateData(Model):
    """
    Instantiates a template as a python type. To customize the class,
    add it to the ``classes`` key and specify the template type.

    Code to be wrapped:

    .. code-block:: c++

        template <typename T>
        class MyClass {};

    To bind ``MyClass<int>`` as the python class ``MyIntClass``, add this
    to your YAML:

    .. code-block:: yaml

        classes:
          MyClass:
            template_params:
            - T

        templates:
          MyIntClass:
            qualname: MyClass
            params:
            - int
    """

    #: Fully qualified name of instantiated class
    qualname: str

    #: Template parameters to use
    params: List[str]

    #: If specified, put the template instantiation in a sub.pack.age
    subpackage: Optional[str] = None

    #: Set the docstring for the template instance
    doc: Optional[str] = None

    #: Text to append to the (autoconverted) docstring for the template instance
    doc_append: Optional[str] = None


class Defaults(Model):
    """
    Defaults to apply to everything
    """

    #: Set this to True to ignore functions/enums/classes at namespace scope
    #: by default if they aren't present in the YAML
    ignore: bool = False

    #: If False and default ignore is True, don't report missing items
    report_ignored_missing: bool = True


class AutowrapConfigYaml(Model):
    """
    Format of the file in [tool.robotpy-build.wrappers."PACKAGENAME"]
    generation_data
    """

    defaults: Defaults = Field(default_factory=Defaults)

    strip_prefixes: List[str] = []

    #: Adds ``#include <FILENAME>`` directives to the top of the autogenerated
    #: C++ file, after autodetected include dependencies are inserted.
    extra_includes: List[str] = []

    #: Adds ``#include <FILENAME>`` directives after robotpy_build.h is
    #: included, but before any autodetected include dependencies. Only use
    #: this when dealing with broken headers.
    extra_includes_first: List[str] = []

    #: Specify raw C++ code that will be inserted at the end of the
    #: autogenerated file, inside a function. This is useful for extending
    #: your classes or providing other customizations. The following C++
    #: variables are available:
    #:
    #: * ``m`` is the ``py::module`` instance
    #: * ``cls_CLASSNAME`` are ``py::class`` instances
    #: * ... lots of other things too
    #:
    #: The trampoline class (useful for accessing protected items) is available
    #: at ``{CLASSNAME}_Trampoline``
    #:
    #: To see the full list, run a build and look at the generated code at
    #: ``build/*/gensrc/**/*.cpp``
    #:
    #: Recommend that you use the YAML multiline syntax to specify it:
    #:
    #: .. code-block:: yaml
    #:
    #:    inline_code: |
    #:      cls_CLASSNAME.def("get42", []() { return 42; });
    inline_code: Optional[str] = None

    #: Key is the attribute (variable) name
    #:
    #: .. code-block:: yaml
    #:
    #:    attributes:
    #:      my_variable:
    #:        # customizations here, see PropData
    #:
    attributes: Dict[str, PropData] = {}

    #: Key is the class name
    #:
    #: .. code-block:: yaml
    #:
    #:    classes:
    #:      CLASSNAME:
    #:        # customizations here, see ClassData
    #:
    classes: Dict[str, ClassData] = {}

    #: Key is the function name
    #:
    #: .. code-block:: yaml
    #:
    #:    functions:
    #:      fn_name:
    #:        # customizations here, see FunctionData
    #:
    functions: Dict[str, FunctionData] = {}

    #: Key is the enum name, for enums at global scope
    #:
    #: .. code-block:: yaml
    #:
    #:    enums:
    #:      MyEnum:
    #:        # customizations here, see EnumData
    #:
    enums: Dict[str, EnumData] = {}

    #: Instantiates a template. Key is the name to give to the Python type.
    #:
    #: .. code-block:: yaml
    #:
    #:    templates:
    #:      ClassName:
    #:        # customizations here, see TemplateData
    #:
    templates: Dict[str, TemplateData] = {}

    #: Extra 'using' directives to insert into the trampoline and the
    #: wrapping scope
    typealias: List[str] = []

    #: Encoding to use when opening this header file
    encoding: str = "utf-8-sig"

    @validator("attributes", pre=True)
    def validate_attributes(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = PropData()
        return value

    @validator("classes", pre=True)
    def validate_classes(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = ClassData()
        return value

    @validator("enums", pre=True)
    def validate_enums(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = EnumData()
        return value

    @validator("functions", pre=True)
    def validate_functions(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = FunctionData()
        return value

    @classmethod
    def from_file(cls, fname) -> "AutowrapConfigYaml":
        with open(fname) as fp:
            data = yaml.safe_load(fp)

        if data is None:
            data = {}

        return cls(**data)



from rpytest import ft


def test_ignore_fn():
    assert not hasattr(ft._rpytest_ft, "fnIgnore")
    assert ft.fnIgnoredParam() == 3


def test_ignore_cls():
    assert not hasattr(ft._rpytest_ft, "IgnoredClass")
    assert not hasattr(ft._rpytest_ft, "IgnoredClassWithEnum")

    c = ft.ClassWithIgnored()

    assert not hasattr(c, "fnIgnore")
    assert c.fnIgnoredParam(y=4) == 42 + 4

    assert not hasattr(c, "ignoredProp")


def test_ignore_cls_enum():
    assert not hasattr(ft.ClassWithIgnored, "IgnoredInnerEnum")
    assert not hasattr(ft.ClassWithIgnored.InnerEnumWithIgnored, "Param1")
    assert ft.ClassWithIgnored.InnerEnumWithIgnored.Param2 == 2


def test_ignored_enums():
    assert not hasattr(ft._rpytest_ft, "IgnoredEnum")

    assert not hasattr(ft.EnumWithIgnored, "Ignored")
    assert ft.EnumWithIgnored.NotIgnored == 1


#
# ignored_by_default
#


def test_ignored_by_default_fn():
    assert not hasattr(ft._rpytest_ft, "id_fnIgnore")
    assert ft._rpytest_ft.id_fnEnable() == 2


def test_ignored_by_default_enum():
    assert not hasattr(ft._rpytest_ft, "id_IgnoredEnum")
    assert ft._rpytest_ft.id_EnabledEnum.Param3 == 3


def test_ignored_by_default_class():
    assert not hasattr(ft._rpytest_ft, "id_IgnoreClass")
    o = ft._rpytest_ft.id_EnabledClass()
    assert o.fn() == 3
    assert o.fn_missing() == 4
    assert ft._rpytest_ft.id_EnabledClass.InnerEnum.Param6 == 6
    assert ft._rpytest_ft.id_EnabledClass.InnerEnumMissing.Param7 == 7



from rpytest import ft
import pytest
import re

#
# main module
#


def test_raise_from():
    with pytest.raises(ValueError) as excinfo:
        ft.raise_from()
    assert str(excinfo.value) == "outer"
    assert str(excinfo.value.__cause__) == "inner"


def test_raise_from_already_set():
    with pytest.raises(ValueError) as excinfo:
        ft.raise_from_already_set()
    assert str(excinfo.value) == "outer"
    assert str(excinfo.value.__cause__) == "inner"


#
# abstract.h
#


class MyBadAbstract(ft.Abstract):
    pass


def test_bad_abstract():
    m = MyBadAbstract()

    with pytest.raises(RuntimeError):
        m.mustOverrideMe()


class MyGoodAbstract(ft.Abstract):
    def mustOverrideMe(self):
        return 0x3


def test_good_abstract():
    m = MyGoodAbstract()
    assert m.mustOverrideMe() == 0x3


class MyBadPrivateAbstract(ft.PrivateAbstract):
    pass


def test_private_abstract():
    # it's private, you can't call it
    assert not hasattr(ft.PrivateAbstract, "_mustOverrideMe")


def test_bad_private_abstract():
    m = MyBadPrivateAbstract()

    with pytest.raises(
        RuntimeError,
        match=r".*"
        + re.escape(
            'does not override required function "PrivateAbstract::_mustOverrideMe"'
        ),
    ):
        ft.PrivateAbstract.getPrivateOverride(m)


class MyGoodPrivateAbstract(ft.PrivateAbstract):
    def _mustOverrideMe(self):
        return 0x3


def test_good_private_abstract():
    m = MyGoodPrivateAbstract()
    assert m._mustOverrideMe() == 0x3
    assert ft.PrivateAbstract.getPrivateOverride(m) == 0x3


#
# buffers.h
#


def test_buffers():
    o = ft.Buffers()
    o.set_buffer(b"12345")

    b = bytearray(4)
    l = o.get_buffer1(b)
    assert b == b"1234"
    assert l == 4

    b = bytearray(4)
    l = o.get_buffer2(b)
    assert b == b"1234"
    assert l == 4

    bi = b"2345"
    bo = bytearray(4)
    l = o.inout_buffer(bi, bo)
    assert bo == b"3456"


def test_buffers_v():
    o = ft.Buffers()
    o.v_set_buffer(b"12345")

    b = bytearray(4)
    l = o.v_get_buffer1(b)
    assert b == b"1234"
    assert l == 4

    b = bytearray(4)
    l = o.v_get_buffer2(b)
    assert b == b"1234"
    assert l == 4


#
# factory.h
#


def test_factory():
    o = ft.HasFactory(4)
    assert o.m_x == 5


#
# gilsafe_container.h
#


def test_gilsafe_container():
    ft.GilsafeContainer.check()


#
# inline_code.h
#


def test_inline_code():
    o = ft.InlineCode()
    assert o.get2() == 2
    assert o.get4() == 4


def test_cpp_code_with_constant():
    o = ft.InlineCode()
    assert o.cpp_code_with_constant() == 4


#
# ns_class.h
#


def test_ns_class():
    assert ft._rpytest_ft.NSClass().getN() == 4


#
# operators.h
#


def test_operators_eq():
    o1 = ft.HasOperator(1)
    o1a = ft.HasOperator(1)
    o2 = ft.HasOperator(2)

    assert o1 == o1a
    assert not (o1 == o2)
    assert o1 != o2


def test_operators_eq2():
    o1 = ft.HasOperatorNoDefault(1)
    o1a = ft.HasOperatorNoDefault(1)
    o2 = ft.HasOperatorNoDefault(2)

    assert o1 == o1a
    assert not (o1 == o2)
    assert o1 != o2


#
# static_only.h
#


def test_static_only():
    # shouldn't be able to construct
    with pytest.raises(TypeError):
        ft.StaticOnly()

    # should be able to call static
    assert ft.StaticOnly.callme() == 0x56


#
# using.h / using2.h
#


def test_using_fwddecl():
    f = ft.FwdDecl()
    f.x = 42
    u = ft.Using4()
    assert u.getX(f) == 43


#
# virtual_xform.h
#
# Check to see that python and C++ see different things for each type
# - py: () -> str
# - C++ (stringstream&) -> void
#


def test_virtual_xform():
    base = ft.VBase()

    with pytest.raises(RuntimeError):
        assert base.pure_io() == "ohai"

    with pytest.raises(RuntimeError):
        assert ft.check_pure_io(base)

    assert base.impure_io() == "py vbase impure + c++ vbase impure"
    assert ft.check_impure_io(base) == "c++ vbase impure"

    assert base.different_cpp_and_py(1) == 3
    assert ft.check_different_cpp_and_py(base, 1) == 2

    class PyChild(ft.VBase):
        def pure_io(self) -> str:
            return "pychild pure"

        def impure_io(self) -> str:
            return "pychild impure"

    pychild = PyChild()
    assert pychild.pure_io() == "pychild pure"
    assert ft.check_pure_io(pychild) == "vbase-xform-pure pychild pure"

    assert pychild.impure_io() == "pychild impure"
    assert ft.check_impure_io(pychild) == "vbase-xform-impure pychild impure"

    child = ft.VChild()

    assert child.pure_io() == "py vchild pure + c++ vchild pure"
    assert ft.check_pure_io(child) == "c++ vchild pure"

    assert child.impure_io() == "py vchild impure + c++ vchild impure"
    assert ft.check_impure_io(child) == "c++ vchild impure"

    class PyGrandChild(ft.VChild):
        def pure_io(self) -> str:
            return "pygrandchild pure"

        def impure_io(self) -> str:
            return "pygrandchild impure"

    pygrandchild = PyGrandChild()

    assert pygrandchild.pure_io() == "pygrandchild pure"
    assert ft.check_pure_io(pygrandchild) == "vchild-xform-pure pygrandchild pure"

    assert pygrandchild.impure_io() == "pygrandchild impure"
    assert ft.check_impure_io(pygrandchild) == "vchild-xform-impure pygrandchild impure"


#
# Misc
#


# ensure that not calling __init__ from a inherited class raises TypeError
def test_init_raises():
    called = [False]

    class PyGoodAbstract(ft.Abstract):
        def __init__(self):
            called[0] = True

    with pytest.raises(TypeError):
        PyGoodAbstract()
    assert called == [True]


def test_subpkg():
    from rpytest.ft.subpkg import SPClass
